import java.util.*;

class Main {
    public static void main(String[] args) {
        List list = new SimpleLinkedList();

        // Test the list.
        list.add("dog");
        list.add("cat");
        list.remove("dog");
        list.add("pig");
        System.out.println( list );    // [cat, pig]
        list.set(list.size()-1, "dog");
        System.out.println( list );    // [cat, dog]

        // Traverse the list backwards
        for (ListIterator it=list.listIterator(list.size()); 
            it.hasPrevious(); ) {
            System.out.println( it.previous() ); 
        }    // dog cat

        // Modify the list while an iterator is active.
        Iterator it = list.iterator();
        list.remove("cat");
        it.next();                     // ConcurrentModificationException
    }
}

class SimpleLinkedList extends AbstractSequentialList {
    private Node header = new Node(null, null, null);
    private int size = 0;

    public SimpleLinkedList() {
        header.next = header.previous = header;
    }

    public SimpleLinkedList(List l) {
        super();
        addAll(l);
    }

    public ListIterator listIterator(int index) {
        return new OurListIterator(index);
    }

    public int size() {
        return size;
    }

    private static class Node {
        Node(Node next, Node previous, Object element) {
            this.next = next;
            this.previous = previous;
            this.element = element;
        }
        Node next;
        Node previous;
        Object element;
    }

    private class OurListIterator implements ListIterator {
        private Node next;
        private int nextIndex;
        private Node lastReturned = null;
        private int expectedModCount = getModCount();

        OurListIterator(int index) {
            if (index < 0 || index > size) {
                throw new IndexOutOfBoundsException("Index: " + index);
            }

            // Iterate from front or back, whichever is closer
            if (index < size/2) {
                next = header.next;
                for (nextIndex=0; nextIndex<index; nextIndex++) {
                    next = next.next;
                }
            } else {
                next = header;
                for (nextIndex=size; nextIndex>index; nextIndex--) {
                    next = next.previous;
                }
            }
        }
        public void add(Object o) {
            checkForComodification();
            Node newNode = new Node(next, next.previous, o);
            next.previous = newNode;
            newNode.previous.next = newNode;
            size++;
            nextIndex++;
            lastReturned = null;
            incModCount();
            expectedModCount++;
        }

        public boolean hasNext() {
            return nextIndex < size;
        }

        public boolean hasPrevious() {
            return nextIndex > 0;
        }

        public Object next() {
            checkForComodification();
            if (!hasNext()) {
                throw new NoSuchElementException();
            }
            lastReturned = next;
            next = next.next;
            nextIndex++;
            return lastReturned.element;
        }

        public int nextIndex() {
            checkForComodification();
            return nextIndex;
        }

        public Object previous() {
            checkForComodification();
            if (!hasPrevious()) {
                throw new NoSuchElementException();
            }
            next = lastReturned = next.previous;
            nextIndex--;
            return lastReturned.element;
        }

        public int previousIndex() {
            checkForComodification();
            return nextIndex-1;
        }

        public void remove() {
            if (lastReturned==null) {
                throw new IllegalStateException();
            }
            checkForComodification();
            if (next==lastReturned) {
                next = lastReturned.next;
            } else {
                nextIndex--;
            }
            lastReturned.previous.next = lastReturned.next;
            lastReturned.next.previous = lastReturned.previous;
            lastReturned = null;
            size--;
            incModCount();
            expectedModCount++;
        }

        public void set(Object o) {
            if (lastReturned==null) {
                throw new IllegalStateException();
            }
            checkForComodification();
            lastReturned.element = o;
        }

        private void checkForComodification() {
            if (getModCount() != expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }
    }

    // "Bridge methods" required to give inner class access to protected field
    private int getModCount() {
        return modCount;
    }
    private void incModCount() {    
        modCount++;
    }
}
